#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

# Import required libraries
require 'resolv'
require 'rest-client'
require 'json'
require 'time'
require 'logger'
require 'active_record'
require 'zk'
require 'set'

# Class for application logging
class AppLogger
  def self.logger
    @logger ||= Logger.new(STDOUT).tap { |log| log.level = Logger::DEBUG }
  end
end

class Formatter
  def self.break_line
    "-" * 80 + "\n"
  end
end

# Main class responsible for asset discovery
class InventoryDeviceFlowDiscovery
  # Retrieves the Druid broker URL
  def self.druid_url
    @druid_url ||= retrieve_druid_broker_url
  end

  # Fetches the Druid broker URL from Zookeeper
  def self.retrieve_druid_broker_url
    zk_data = YAML.load(File.read('config/rbdruid_config.yml'))[Rails.env].symbolize_keys
    zk_connect = zk_data[:zk_connect]
    discovery_path = '/druid/discoveryPath/druid:broker'

    brokers = fetch_brokers(zk_connect, discovery_path)
    raise "No active Druid Broker found." if brokers.empty?

    brokers.first
  end

  # Queries Zookeeper for active brokers
  def self.fetch_brokers(zk_connect, discovery_path)
    brokers_names = []

    ZK.open(zk_connect) do |zk|
      zk.children(discovery_path).shuffle.each do |broker|
        zktdata = YAML.load(zk.get("#{discovery_path}/#{broker}").first)
        if zktdata["address"] && zktdata["port"]
          brokers_names << "http://#{zktdata['address']}:#{zktdata['port']}/druid/v2/?pretty"
        end
      end
    end

    brokers_names
  end

  def initialize
    @messages_printed = Hash.new(false)
    @result = ""
  end

  # Main method to execute the asset inventory job
  def execute
    @result += "Starting Assets Inventory Job...\n"
    Delayed::Worker.logger.info "Starting Assets Inventory Job.."

    conn = establish_db_connection

    if conn
      Delayed::Worker.logger.info "Fetching data_sources.."
      data_sources = fetch_data_sources_from_sensor
      
      Delayed::Worker.logger.info "Start process_data.."
      total_count = process_data(conn, data_sources)

      @result += "A total of #{total_count} records (MACs and IPs) have been processed.\n"
    else
      @result += "No database connection was established.\n"
    end

    # Closes the database connection if it was established
    conn.close if conn
    @result
  rescue StandardError => e
    AppLogger.logger.error("Error: #{e.message}")
    AppLogger.logger.error(e.backtrace.join("\n"))
    "Error: #{e.message}"
  end

  private

  # Establishes a connection to the PostgreSQL database.
  def establish_db_connection
    ActiveRecord::Base.connection_pool.with_connection do
      @result += "Connection to PostgreSQL established successfully.\n"
      ActiveRecord::Base.connection
    end
  rescue ActiveRecord::ConnectionNotEstablished => e
    @result += "Failed to connect to PostgreSQL: #{e.message}"
    nil
  end

  # Fetches a list of data sources from the 'Sensor' model.
  def fetch_data_sources_from_sensor
    sensors = Sensor.where(domain_type: Sensor::NAMESPACE).pluck(:uuid, :name)
    sensors.map { |uuid, name| ["rb_flow_#{uuid}", name] }.unshift(["rb_flow", "root"])
  end

  # Processes data by iterating through a list of data sources, fetching data, and recording results.
  def process_data(conn, data_sources)
    start_time, end_time = calculate_time_interval
    @result += "Time range for the query: #{start_time} to #{end_time}\n"
    @result += Formatter.break_line

    total_mac_insert_or_udpate_count = 0, 0

    all_mac_addresses = []

    flow_sensors_uuids_with_homenets = Sensor.where(type: [Sensor::FLOW, Sensor::MERAKI]).select{|s| s.homenets && s.homenets.any? }.map(&:uuid) rescue []

    data_sources.each do |data_source, name|
      AppLogger.logger.debug("Searching in domain: #{name}\n")
      Delayed::Worker.logger.info "Fetching and process data for #{name}\n"

      # TODO: optimize this to loop only over sensors that are inside this particular data_source
      flow_sensors_uuids_with_homenets.each do |sensor_uuid|
        Delayed::Worker.logger.info "Fetching sensor #{sensor_uuid}\n"
        responses = fetch_and_process_data(data_source, start_time, end_time, sensor_uuid)

        next unless responses

        Delayed::Worker.logger.info "Calling process_responses.."
        process_responses(responses, all_mac_addresses, name)
      end
    end

    # Before inserting or updating the MAC records, remove duplicates
    all_mac_addresses.uniq!

    Delayed::Worker.logger.info "Number of mac to be insert/update: #{all_mac_addresses.count}"
    total_mac_insert_or_udpate_count = insert_or_update_mac_records(conn, all_mac_addresses)

    Delayed::Worker.logger.info "Total MACs inserted or updated: #{total_mac_insert_or_udpate_count}"
    @result += "Total MACs inserted or updated: #{total_mac_insert_or_udpate_count}\n"
    @result += Formatter.break_line

    total_mac_insert_or_udpate_count
  end

  # Processes the responses fetched from the data sources.
  def process_responses(responses, all_mac_addresses, organization_uuid)
    flow_sensors_uuids_with_homenets = Sensor.where(type: [Sensor::FLOW, Sensor::MERAKI]).select{|s| s.homenets && s.homenets.any? }.map(&:uuid) rescue []
    Delayed::Worker.logger.info "Flow sensors with homenets: #{flow_sensors_uuids_with_homenets}"
    
    responses.each do |dimension, response_json|
      response_json.each do |item|
        client_mac = item.dig('event', 'client_mac')

        next unless client_mac

        sensor_uuid = item.dig('event', 'sensor_uuid')

        next unless sensor_uuid && flow_sensors_uuids_with_homenets.include?(sensor_uuid)

        client_mac_vendor = item.dig('event', 'client_mac_vendor')
        lan_ip = item.dig('event', 'lan_ip')
        wan_ip = item.dig('event', 'wan_ip')

        next unless lan_ip && wan_ip

        ip_address = private_ip?(lan_ip) ? lan_ip : wan_ip

        # Extract multiple UUIDs
        uuids = {
          sensor_uuid: item.dig('event', 'sensor_uuid'),
          organization_uuid: item.dig('event', 'organization_uuid'),
          campus_uuid: item.dig('event', 'campus_uuid'),
          building_uuid: item.dig('event', 'building_uuid'),
          service_provider_uuid: item.dig('event', 'service_provider_uuid'),
          floor_uuid: item.dig('event', 'floor_uuid'),
          namespace_uuid: item.dig('event', 'namespace_uuid'),
          market_uuid: item.dig('event', 'market_uuid'),
          deployment_uuid: item.dig('event', 'deployment_uuid')
        }

        all_mac_addresses << { mac_address: client_mac, mac_vendor: client_mac_vendor, ip_address: ip_address, uuids: uuids }
      end
    end
  end

  # Inserts the MAC records into the database.
  def insert_or_update_mac_records(conn, mac_data)
    count = 0
    updated_devices_ids = []
    mac_data.each do |data|
      mac_address = data[:mac_address]
      ip_address = data[:ip_address]
      mac_vendor = data[:mac_vendor]
      uuids = data[:uuids]
      device_name = resolve_ip(ip_address)
      count += 1 if insert_or_update_record(conn, mac_address, mac_vendor, ip_address, device_name, uuids, updated_devices_ids)
    end
    count
  end

  # Inserts a record into the database for a given value (e.g., MAC or IP address),
  # ensuring it is associated with the nearest sensor UUID based on hierarchy.
  def insert_or_update_record(conn, mac_address, mac_vendor, ip_address, device_name, uuids, updated_devices_ids)
    # Check if the value is '00:00:00:00:00:00' (invalid MAC address)
    # If true, the method returns false and doesn't proceed with the insertion.
    return false if mac_address.strip == '00:00:00:00:00:00'
    
    # Retrieves the nearest sensor UUID from the provided UUIDs hash
    nearest_id = get_nearest_sensor_id(uuids)

    return false unless nearest_id

    # Check if a record with the same value and sensor already exists in the database.
    # If it exists, return false to avoid duplicate insertion.
    mac_object = MacObject.find_by("value = ? AND sensor_id = ?", mac_address, nearest_id)
    device = mac_object.inventory_device if mac_object

    # Update mac_object if needed
    if mac_object
      mac_object_changed = false
      if mac_object.value != mac_address
        mac_object.value = mac_address
        mac_object_changed = true
      end

      if mac_object.vendor != mac_vendor
        mac_object.vendor = mac_vendor
        mac_boject_changed = true
      end

      if mac_object_changed
        mac_object.updated_at = Time.current
        Delayed::Worker.logger.info "Updating MAC object: #{mac_object.value}"
        mac_object.save
      end
    end

    if device
      updated = false

      unless device.locked && device.ips.include?(ip_address)
        # Clear old list of ip addresses (on the first ip update)
        device.ips = [] unless updated_devices_ids.include?(device.id)

        device.ips |= [ip_address]
        updated = true
      end

      unless device.locked && device.mac_objects.include?(mac_object)
        device.mac_objects |= [mac_object]
        updated = true
      end

      unless device.sources.include?('Traffic')
        device.sources |= ['Traffic']
        updated = true
      end

      if !device.locked && device.inventory_device_type_object.nil?
        inventory_device_type_object = DeviceTypeClassifier.classify(device)
        if inventory_device_type_object
          device.inventory_device_type_object = inventory_device_type_object
          updated = true
        end
      end

      if !device.locked && !device.name.present? && device_name.present? && device_name != device.name
        device.name = device_name
        updated = true
      end

      # This avoid to update last_seen multiple events in same execute run
      unless updated_devices_ids.include?(device.id)
        device.last_seen = Time.current
        updated = true
      end

      if updated
        device.updated_at = Time.current
        updated_devices_ids << device.id
        Delayed::Worker.logger.info "Updating InventoryDevice: #{device.id}"
        return device.save!
      else
        return false
      end
    end

    Delayed::Worker.logger.info "Inserting new MacObject and InventoryDevice: #{mac_address}"

    MacObject.transaction do
      # Create a new instance of the model_class (e.g., MacObject or NetObject)
      # Populate its fields with the value, resolved name, and the nearest sensor ID.
      mac_object = MacObject.new(
        value: mac_address,
        name: (device_name || mac_address),
        vendor: mac_vendor,
        user_id: nil,                  # No user_id is assigned
        created_at: Time.current,      # Set the current time for the creation timestamp
        updated_at: Time.current,      # Set the current time for the updated timestamp
        sensor_id: nearest_id           # Associate the record with the nearest sensor ID
      )
      # Save the new object to the database
      mac_object.save!
    end

    # Begin a database transaction to safely insert the new record.
    # This ensures that if anything goes wrong during the insertion, the entire operation can be rolled back.
    InventoryDevice.transaction do
      # Create a new instance of the model_class (e.g., MacObject or NetObject)
      # Populate its fields with the value, resolved name, and the nearest sensor ID.
      object = InventoryDevice.new(
        mac_objects: [mac_object],
        ips: [ip_address],
        name: device_name,              # Resolved name for the value (e.g., human-readable MAC or IP name)
        sources: ['Traffic'],           # Set the source of the record to 'Traffic'
        created_at: Time.current,       # Set the current time for the creation timestamp
        first_seen: Time.current,       # Set the current time for the first_seen timestamp
        last_seen: Time.current,        # Set the current time for the last_seen timestamp
        updated_at: Time.current,       # Set the current time for the updated timestamp
        sensor_id: nearest_id           # Associate the record with the nearest sensor ID
      )
      # Save the new object to the database
      if object.save!
        # Classify object
        inventory_device_type_object = DeviceTypeClassifier.classify(object)
        if inventory_device_type_object
          object.inventory_device_type_object = inventory_device_type_object
          return object.save!
        else
          return true
        end
      end
    end

    false
  end

  # Helper method to get the nearest sensor ID based on hierarchy
  def get_nearest_sensor_id(uuids)
    # Filter out nil UUIDs and put them in a list
    uuid_list = [
      uuids[:floor_uuid],
      uuids[:building_uuid],
      uuids[:campus_uuid],
      uuids[:organization_uuid],
      uuids[:service_provider_uuid],
      uuids[:market_uuid],
      uuids[:deployment_uuid],
      uuids[:sensor_uuid]
    ].compact

    # If there are no UUIDs in the list, return nil
    return nil if uuid_list.empty?

    # Find the nearest UUID based on the smallest `depth`
    nearest_id = Sensor.where(uuid: uuid_list).order(:depth).pluck(:uuid).first

    # If no UUID is found, return nil
    return nil unless nearest_id

    # Find the sensor corresponding to the nearest UUID
    sensor = Sensor.find_by(uuid: nearest_id)

    # If the sensor has no parent_id or the parent doesn't exist, return the sensor ID
    return sensor.id if sensor && sensor.is_domain?

    # Find the parent sensor
    parent_sensor = Sensor.find_by(id: sensor.parent_id)

    if parent_sensor.is_client_proxy?
      return nil unless parent_sensor.parent_id

      # If the parent sensor is a proxy, find the grandparent sensor
      parent_sensor = Sensor.find_by(id: parent_sensor.parent_id)
    end

    return nil unless parent_sensor

    # If the parent sensor exists, return its ID; otherwise, return nil
    parent_sensor ? parent_sensor.id : nil
  end

  def private_ip?(ip)
    return false if ip.nil?
  
    private_ip_regex = /^(10\.|172\.(1[6-9]|2\d|3[01])\.|192\.168\.|fc[0-9a-fA-F]{2}:|fd[0-9a-fA-F]{2}:)/
    ip.match?(private_ip_regex)
  end

  # Main function that fetches and processes data from a source.
  def fetch_and_process_data(data_source, start_time, end_time, sensor_uuid)
    # Define the dimensions to query
    query_responses = {}

    # Build the query using the build_query function.
    query = build_query(data_source, start_time, end_time, sensor_uuid)
    druid_url = InventoryDeviceFlowDiscovery.druid_url # Get the URL for the Druid service.

    begin
      # Perform the HTTP POST request to Druid with the query in JSON format.
      response = RestClient.post(druid_url, query.to_json, { content_type: :json })
      parsed_response = JSON.parse(response.body) # Parse the response into JSON.

      # Save the response only if it's not empty.
      query_responses[:client_mac] = parsed_response unless parsed_response.empty?
    rescue RestClient::ExceptionWithResponse, JSON::ParserError => e
      # Log the error if one occurs.
      AppLogger.logger.error("Error querying Druid: #{e.message}")
      AppLogger.logger.error(e.response.body) if e.respond_to?(:response) # Log the response body if available.
      raise # Re-raise the exception.
    end

    query_responses # Return the responses obtained for each dimension.
  end

  # Function that builds the query for Druid with the given parameters.
  def build_query(data_source, start_time, end_time, sensor_uuid)
    {
      queryType: "groupBy", # Type of query that groups results.
      dataSource: data_source, # Specifies the data source.
      granularity: "all", # Aggregation without granularity (all data).
      dimensions: [
        "client_mac", "client_mac_vendor", "lan_ip", "wan_ip", "sensor_uuid", "service_provider_uuid",
        "namespace_uuid", "organization_uuid", "campus_uuid", "building_uuid",
        "floor_uuid", "market_uuid", "deployment_uuid"
      ], # List of dimensions to include in the results.
      context: {
        timeout: 90000, # Maximum wait time in milliseconds (90 seconds).
        skipEmptyBuckets: true # Skip empty results.
      },
      limitSpec: {
        type: "default", # Type of result limitation.
        limit: 50000, # Maximum number of results to return.
        columns: [
          { dimension: "client_mac", direction: "ascending" } # Sort by client_mac in ascending order.
        ]
      },
      intervals: ["#{start_time}/#{end_time}"], # Time interval for the query.
      filter: {
        type: "and",
        fields: [
          {
            type: "or",
            fields: [
              {
                type: "regex",
                dimension: "lan_ip",
                pattern: "^(10\\.|172\\.(1[6-9]|2\\d|3[01])\\.|192\\.168\\.|fc[0-9a-fA-F]{2}:|fd[0-9a-fA-F]{2}:)"
              },
              {
                type: "regex",
                dimension: "wan_ip",
                pattern: "^(10\\.|172\\.(1[6-9]|2\\d|3[01])\\.|192\\.168\\.|fc[0-9a-fA-F]{2}:|fd[0-9a-fA-F]{2}:)"
              }
            ]
          },
          {
            type: "selector",
            dimension: "sensor_uuid",
            value: "#{sensor_uuid}"
          },
          {
            type: "not",
            field: {
              type: "selector",
              dimension: "client_mac",
              value: "00:00:00:00:00:00"
            }
          }
        ]
      }
    }
  end

  # Function that calculates the time interval (last hour).
  # Returns two values: the start time (one hour ago) and the end time (now).
  def calculate_time_interval
    now = Time.now # Current time.
    start_time = (now - 3600).utc.iso8601 # Calculate the start time by subtracting one hour.
    end_time = now.utc.iso8601 # End time (current).
    [start_time, end_time] # Return the time interval.
  end

  # Function that attempts to resolve an IP address to its domain name.
  # If it cannot be resolved, it returns the original IP.
  def resolve_ip(ip)
    # Delayed::Worker.logger.info("Resolving IP: #{ip}") # Log the IP address being resolved.
    Resolv.getname(ip) # Try to resolve the domain name associated with the IP.
  rescue Resolv::ResolvError
    nil # If an error occurs while resolving, return the IP unchanged.
  end
end
